@@ -8,6 +8,8 @@
 #' @param useDVI A logical, indicating if the DVI section of the fam file should
 #'   be identified and parsed. If `NA` (the default), the DVI section is
 #'   included if it is present in the input file.
+#' @param Xchrom A logical. If TRUE, the `chrom` attribute of all markers will
+#'   be set to "X". (Default = FALSE.)
 #' @param verbose A logical; if TRUE, various information is written to the
 #'   screen during the parsing process.
 #'
@@ -24,7 +26,7 @@
 #'
 #' @importFrom pedmut mutationMatrix
 #' @export
-readFam = function(famfile, useDVI = NA, verbose = TRUE) {
+readFam = function(famfile, useDVI = NA, Xchrom = FALSE, verbose = TRUE) {
   if(!endsWith(famfile, ".fam"))
     stop("Input file must end with '.fam'", call. = FALSE)
 
@@ -92,6 +94,9 @@ readFam = function(famfile, useDVI = NA, verbose = TRUE) {
 
   ### Fixed relations
 
+  # Storage for twins
+  twins = list()
+
   kr.line = id.line
   if(x[kr.line] != "Known relations")
     stop2(sprintf('Expected line %d to be "Known relations", but found: "%s"', id.line, x[id.line]))
@@ -149,12 +154,24 @@ readFam = function(famfile, useDVI = NA, verbose = TRUE) {
       fidx.i = c(fidx, integer(nFem.i + nMal.i))
       midx.i = c(midx, integer(nFem.i + nMal.i))
 
+      # Print summary
+      if(verbose)
+        message(sprintf(" Pedigree '%s' (%d extra females, %d extra males)", ped.name, nFem.i, nMal.i))
+
       # Add fixed relations
       nRel.i = as.integer(x[ped.line + 4])
       rel.line = ped.line + 5
       for(i in seq_len(nRel.i)) {
         par.idx = as.integer(x[rel.line]) + 1
         child.idx = as.integer(x[rel.line+1]) + 1
+        if(is.na(par.idx)) {
+          if(grepl("Direct", x[rel.line])) {
+            par.idx = as.integer(substring(x[rel.line], 1, 1)) + 1
+            twins = c(twins, list(par.idx, child.idx))
+            if(verbose) message("  Twins: ", toString(id[c(par.idx, child.idx)]))
+            stop2("File contains twins - this is not supported yet")
+          }
+        }
         if(sex[par.idx] == 1)
           fidx.i[child.idx] = par.idx
         else
@@ -163,9 +180,6 @@ readFam = function(famfile, useDVI = NA, verbose = TRUE) {
         rel.line = rel.line + 2
       }
 
-      # Print summary
-      if(verbose)
-        message(sprintf(" Pedigree '%s': %d extra females, %d extra males", ped.name, nFem.i, nMal.i))
 
       # Convert to familiaspedigree and insert in list
       pedigrees[[ped.idx]] = asFamiliasPedigree(id.i, fidx.i, midx.i, sex.i)
@@ -300,26 +314,25 @@ readFam = function(famfile, useDVI = NA, verbose = TRUE) {
     if(is.na(dvi.start))
       stop2("Expected keyword '[DVI]' not found")
     dvi.lines = raw[dvi.start:length(raw)]
-    dvi.families = readDVI(dvi.lines)
+    dvi.families = readDVI(dvi.lines, verbose = verbose)
 
     if(verbose)
-      message("Returning the following families:")
-
-    famnames = names(dvi.families)
-    for(i in seq_along(dvi.families)) {
-      fam = dvi.families[[i]]
-      if(verbose) {
-        message(sprintf("%s: %d pedigrees", famnames[i], length(fam$pedigrees)))
-        for(nm in names(fam$pedigrees))
-          message("  ", nm)
-      }
-    }
+      message("*** Finished DVI section ***\n")
 
+    if(verbose)
+      message("\nConverting to `ped` format")
     res = lapply(dvi.families, function(fam) {
       Familias2ped(familiasped = fam$pedigrees, datamatrix = fam$datamatrix,
                    loci = loci, matchLoci = TRUE)
     })
 
+    # Set all chrom attributes to X if indicated
+    if(Xchrom) {
+      if(verbose) message("Changing all chromosome attributes to `X`")
+      chrom(res, seq_along(loci)) = "X"
+    }
+
+    if(verbose) message("")
     return(res)
   }
 
@@ -355,16 +368,27 @@ readFam = function(famfile, useDVI = NA, verbose = TRUE) {
       datamatrix[, 2*i]     = als.i[dm.a2.idx[, i]]
     }
   }
+
+  # Return
   if(!is.null(pedigrees)) {
     if(verbose)
-      message("\nReturning pedigrees with attached database.\n")
-    Familias2ped(familiasped = pedigrees, datamatrix = datamatrix, loci = loci)
+      message("\nConverting to `ped` format")
+    res = Familias2ped(familiasped = pedigrees, datamatrix = datamatrix, loci = loci)
+
+    # Set all chrom attributes to X if indicated
+    if(Xchrom) {
+      if(verbose) message("Changing all chromosome attributes to `X`")
+      chrom(res, seq_along(loci)) = "X"
+    }
   }
   else {
     if(verbose)
-      message("\nReturning database only.\n")
-    readFamiliasLoci(loci = loci)
+      message("\nReturning database only")
+    res = readFamiliasLoci(loci = loci)
   }
+
+  if(verbose) message("")
+  res
 }
 
 
@@ -388,10 +412,10 @@ asFamiliasPedigree = function(id, findex, mindex, sex) {
 ### Utilities for parsing DVI section ###
 #########################################
 
-readDVI = function(rawlines) {
+readDVI = function(rawlines, verbose = TRUE) {
   r = rawlines
   if(r[1] != "[DVI]")
-    stop("I excepted the first line of DVI part to be '[DVI]', but got '", r[1], "'")
+    stop("Expected the first line of DVI part to be '[DVI]', but got '", r[1], "'")
 
   ### Parse raw lines into nested list named `dvi`
   dvi = list()
@@ -418,36 +442,102 @@ readDVI = function(rawlines) {
       ivec = c(ivec[seq_len(br - 1)], name)
       dvi[[ivec]] = list()
     }
+
   }
 
-  family_list = dvi$DVI$`Reference Families`[-1] # remove 'nFamilies'
-  names(family_list) = sapply(family_list, function(fam) fam[[1]][2])
+  # Initialise output list
+  res = list()
 
-  lapply(family_list, parseFamily)
+  # Unidentified persons, if any
+  un = parseUnidentified(dvi$DVI$`Unidentified persons`, verbose = verbose)
+  if(!is.null(un))
+    res$`Unidentified persons` = un
+
+  # Reference families
+  refs_raw = dvi$DVI$`Reference Families`
+  refs = refs_raw[-1] # remove 'nFamilies'
+  stopifnot((nFam <- length(refs)) == as.integer(refs_raw[[c(1,2)]]))
+  if(verbose)
+    message("\nReference families: ", nFam)
+
+  names(refs) = sapply(refs, function(fam) fam[[1]][2])
+  refs = lapply(refs, parseFamily, verbose = verbose)
+
+  # Return
+  c(res, refs)
 }
 
+parseUnidentified = function(x, verbose = TRUE) {
+  if(length(x) == 0)
+    return(NULL)
+
+  nPers = x[[c(1,2)]]
+  if(verbose)
+    message("Unidentified persons: ", nPers)
+
+  if(nPers == "0")
+    return(NULL)
+
+  x = x[-1]
+
+  ### id and sex
+  id = sapply(x, function(p) getValue(p[[1]], iftag = "Name", NA))
+  sex = sapply(x, function(p) getValue(p[[2]], iftag = "Gender", 0))
+  sex[sex == "Male"] = 1
+  sex[sex == "Female"] = 2
+  s = asFamiliasPedigree(as.character(id), 0, 0, as.integer(sex))
+
+  if(verbose)
+    for(nm in id) message("  ", nm)
+
+  ### datamatrix
+  vecs = lapply(x, function(p) dnaData2vec(p$`DNA data`))
+
+  # Remove NULLs
+  vecs = vecs[!sapply(vecs, is.null)]
+
+  # All column names
+  allnames = unique(unlist(lapply(vecs, names)))
+
+  # Ensure same order in each vector, and fill in NA's
+  vecs_ordered = lapply(vecs, function(v) structure(v[allnames], names = allnames))
+
+  # Bind to matrix
+  datamatrix = do.call(rbind, vecs_ordered)
+  rownames(datamatrix) = id[rownames(datamatrix)]
+
+  ### return
+  list(pedigrees = s, datamatrix = datamatrix)
+}
 
 # Convert a "DVI Family" into a list of `datamatrix` and `pedigrees`
-parseFamily = function(x) {
+parseFamily = function(x, verbose = TRUE) {
 
-  persons_list = x$Persons
-  if(persons_list[[c(1,1)]] == "nPersons")
-    persons_list[[1]] = NULL # remove 'nPersons' entry
+  famname = x[[c(1,2)]]
+  nPers = as.integer(x$Persons[[c(1,2)]])
+  nPeds = as.integer(x$Pedigrees[[c(1,2)]])
 
-  ### pedigrees
+  if(verbose)
+    message(sprintf("  %s (%d persons, %d pedigrees)", famname, nPers, nPeds))
+
+  ### Persons
+  persons_list = x$Persons[-1]
+
+  id = sapply(persons_list, function(p) getValue(p[[1]], iftag = "Name", NA))
+  sex = sapply(persons_list, function(p) getValue(p[[2]], iftag = "Gender", 0))
 
-  # collext id and sex of each person
-  id = sapply(persons_list, function(p) p[[1]][2])
-  sex = sapply(persons_list, function(p) p[[2]][2])
   sex[sex == "Male"] = 1
   sex[sex == "Female"] = 2
   sex = as.integer(sex)
 
-  # build pedigrees defined for the family
+  ### pedigrees
   ped_list = x$Pedigrees[-1] # remove "nPedigrees"
-  names(ped_list) = sapply(ped_list, function(pd) pd[[1]][2])
+  names(ped_list) = sapply(ped_list, function(pd) getValue(pd[[1]], iftag = "Name", NA))
 
   pedigrees = lapply(ped_list, function(pd) {
+    if(verbose)
+      message("    ", pd[[1]][2])
+
     this.id = as.character(id)
     this.sex = sex
 
@@ -460,20 +550,40 @@ parseFamily = function(x) {
                     stringsAsFactors = FALSE)
 
     # Add extra individuals if needed (e.g. "Missing person")
-    if(length(extras <- setdiff(po$child, id))) {
+    extras <- setdiff(c(po$parent, po$child), id)
+    if(length(extras)) {
       this.id = c(this.id, extras)
       this.sex = c(this.sex, rep(0L, length(extras)))
     }
 
-    # Create and populate fidx and midx
-    this.fidx = this.midx = integer(length(this.id))
+    names(this.sex) = this.id
+    parent.sex = this.sex[po$parent]
+
+    # Try to fix parents with undecided sex
+    if(any(parent.sex == 0)) {
+      par.nosex = unique(po$parent[parent.sex == 0])
+      for(p in par.nosex) {
+        chi = po$child[po$parent == p] # children of him/her
+        spou = unique(setdiff(po$parent[po$child %in% chi], p))
+        if(all(this.sex[spou] == 1))
+          this.sex[p] = 2
+        else if(all(this.sex[spou] == 2))
+          this.sex[p] = 1
+        else
+          stop2("Cannot decide sex of this parent: ", p)
+      }
+
+      # Now try again
+      parent.sex = this.sex[po$parent]
+    }
 
     parent.idx = match(po$parent, this.id)
-    par.is.male = this.sex[parent.idx] == 1
-
     child.idx = match(po$child, this.id)
-    this.fidx[child.idx[par.is.male]] = parent.idx[par.is.male]
-    this.midx[child.idx[!par.is.male]] = parent.idx[!par.is.male]
+
+    # Create and populate fidx and midx
+    this.fidx = this.midx = integer(length(this.id))
+    this.fidx[child.idx[parent.sex == 1]] = parent.idx[parent.sex == 1]
+    this.midx[child.idx[parent.sex == 2]] = parent.idx[parent.sex == 2]
 
     # Return
     asFamiliasPedigree(this.id, this.fidx, this.midx, this.sex)
@@ -516,3 +626,7 @@ dnaData2vec = function(x) {
   names(res) = paste(rep(val[idx], each = 2), 1:2, sep = ".")
   res
 }
+
+getValue = function(x, iftag, default) {
+  if(x[1] == iftag) x[2] else default
+}
